/*
 *
 *  Title:
 *	Backup
 *
 *  Decription:
 *	Program to read VMS backup tape
 *
 *  Author:
 *	John Douglas CAREY.
 *	Sven-Ove Westberg    (version 3.0)
 *	Pieter Bowman
 *
 *  Net-addess:
 *	john%monu1.oz@seismo.ARPA
 *	luthcad!sow@enea.UUCP
 *	bowman@math.utah.edu
 *
 *  History:
 *	Version 1.0 - September 1984
 *		Can only read variable length records
 *	Version 1.1
 *		Cleaned up the program from the original hack
 *		Can now read stream files
 *	Version 1.2
 *		Now convert filename from VMS to UNIX
 *			and creates sub-directories
 *	Version 1.3
 *		Works on the Pyramid if SWAP is defined
 *	Version 1.4
 *		Reads files spanning multiple tape blocks
 *	Version 1.5
 *		Always reset reclen = 0 on file open
 *		Now output fixed length records
 *
 *      Version 2.0 - July 1985
 *		VMS Version 4.0 causes a rethink !!
 *		Now use mtio operations instead of opening and closing file
 *		Blocksize now grabed from the label
 *
 *	Version 2.1 - September 1985
 *		Handle variable length records of zero length.
 *
 *	Version 2.2 - July 1986
 *		Handle FORTRAN records of zero length.
 *		Inserted exit(0) at end of program.
 *		Distributed program in aus.sources
 *
 *	Version 2.3 - August 1986
 *		Handle FORTRAN records with record length fields
 *		at the end of a block
 *		Put debug output to a file.
 *		Distributed program in net.sources
 *
 *	Version 3.0 - December 1986
 *		Handle multiple saveset 
 *		Remote tape
 *		Interactive mode
 *		File name selection with meta-characters
 *		Convert ; to : in VMS filenames
 *		Flag for usage of VMS directory structure
 *		Flag for "useless" files  eg. *.exe
 *		Flag for use VMS version in file names
 *		Flag for verbose mode
 *		Flag to list the contents of the tape
 *		Distributed to mod.sources
 *
 *
 *		Changed code to allow reading from save set on disk.
 *		Added -b flag to handle changing block size.
 *
 *
 *  Installation:
 *
 *	Computer Centre
 *	Monash University
 *	Wellington Road
 *	Clayton
 *	Victoria	3168
 *	AUSTRALIA
 *
 */
#include	<stdio.h>
#include	<ctype.h>
#include <errno.h>
#include	<sys/ioctl.h>
#include	<sys/types.h>
#ifdef REMOTE
#include	<local/rmt.h>
#include	<sys/stat.h>
#endif
#include	<sys/mtio.h>
#include	<sys/file.h>
#include	<mp.h>
#include	<time.h>

#ifndef SWAP
#ifdef pyr
#define SWAP
#endif pyr

#ifdef sun
#define SWAP
#endif
#endif

#ifdef ultrix
#undef SWAP
#endif

#define FALSE 0
#define TRUE 1

struct bbh {
	short	bbh_dol_w_size;
	short	bbh_dol_w_opsys;
	short	bbh_dol_w_subsys;
	short	bbh_dol_w_applic;
	long	bbh_dol_l_number;
	char	bbh_dol_t_spare_1[20];
	short	bbh_dol_w_struclev;
	short	bbh_dol_w_volnum;
	long	bbh_dol_l_crc;
	long	bbh_dol_l_blocksize;
	long	bbh_dol_l_flags;
	char	bbh_dol_t_ssname[32];
	short	bbh_dol_w_fid[3];
	short	bbh_dol_w_did[3];
	char	bbh_dol_t_filename[128];
	char	bbh_dol_b_rtype;
	char	bbh_dol_b_rattrib;
	short	bbh_dol_w_rsize;
	char	bbh_dol_b_bktsize;
	char	bbh_dol_b_vfcsize;
	short	bbh_dol_w_maxrec;
	long	bbh_dol_l_filesize;
	char	bbh_dol_t_spare_2[22];
	short	bbh_dol_w_checksum;
} *block_header;

struct brh {
	short	brh_dol_w_rsize;
	short	brh_dol_w_rtype;
	long	brh_dol_l_flags;
	long	brh_dol_l_address;
	long	brh_dol_l_spare;
} *record_header;

/* define record types */

#define	brh_dol_k_null	0
#define	brh_dol_k_summary	1
#define	brh_dol_k_volume	2
#define	brh_dol_k_file	3
#define	brh_dol_k_vbn	4
#define brh_dol_k_physvol	5
#define brh_dol_k_lbn	6
#define	brh_dol_k_fid	7

struct bsa {
	short bsa_dol_w_size;
	short bsa_dol_w_type;
	unsigned char *bsa_dol_t_text;
};

#ifdef	STREAM
char	*def_tapefile = "/dev/rts8";
#else
char	*def_tapefile = "/dev/rmt8";
#endif
char	*tapefile;
int istape;

unsigned char	filename[128];	/* VMS file name */
unsigned char	ufn[256];	/* Unix file name */
int	filesize;
time_t unixTimes[4];

unsigned char	recfmt;		/* record format */

#define FAB_dol_C_UDF	0	/* undefined */
#define FAB_dol_C_FIX	1	/* fixed-length record */
#define FAB_dol_C_VAR	2	/* variable-length record */
#define FAB_dol_C_VFC	3	/* variable-length with fixed-length control record */
#define FAB_dol_C_STM	4	/* RMS stream record (valid only for sequential org) */
#define FAB_dol_C_STMLF	5	/* stream record delimited by LF (sequential org only) */
#define FAB_dol_C_STMCR	6	/* stream record delimited by CR (sequential org only) */
#define FAB_dol_C_MAXRFM	6	/* maximum rfm supported */

char	recatt;		/* record attributes */

#define FAB_dol_V_FTN	0x01	/* FORTRAN carriage control character */
#define FAB_dol_V_CR	0x02	/* line feed - record -carriage return */
#define FAB_dol_V_PRN	0x04	/* print-file carriage control */
#define FAB_dol_V_BLK	0x08	/* records don't cross block boundaries */

#define	FANO	100

#ifdef	pyr
static struct	bsa	file_table[FANO];
#else
struct	bsa	file_table[FANO];
#endif

FILE	*f	= NULL;
int	file_count;
short	reclen;
short	recsize;
int	vfcsize;

#ifdef	NEWD
FILE	*lf;
#endif	NEWD

int	fd;		/* tape file descriptor */
#ifdef DEBUG
int 	debugFlag;
#endif /* DEBUG */
int 	bflag, cflag, dflag, eflag, sflag, tflag, vflag, wflag, xflag;
int	setnr = 0;
char	**gargv;
int 	goptind, gargc;

#define	LABEL_SIZE	80
char	label[LABEL_SIZE];

char	*block;
int	blocksize;

struct	mtop	op;

void
CloseFile()
{
  time_t t2[2];

  if (f != NULL) {
    fclose(f);
    f = NULL;

    t2[0] = 0;
    t2[1] = unixTimes[1];	/* Use the revised date from VMS */
				/* for the Unix modify date. */
    utime(ufn, t2);
  }

  file_count = 0;
  reclen = 0;
}

int
typecmp(str)    /* Compare the filename type in str with our list
                   of file type to be ignored.  Return 0 if the
                   file is to be ignored, return 1 if the
                   file is not in our list and should not be ignored. */
register char   *str;
{
        static char *type[] = {
                "exe",          /* vms executable image */
                "lib",          /* vms object library */
                "obj",          /* rsx object file */
                "odl",          /* rsx overlay description file */
                "olb",          /* rsx object library */
                "pmd",          /* rsx post mortem dump */
                "stb",          /* rsx symbol table */
                "sys",          /* rsx bootable system image */
                "tsk",          /* rsx executable image */
		"dir",
		"upd",
		"tlo",
		"tlb",
                ""              /* null string terminates list */
        };
        register int    i;

        i = -1;
        while (*type[++i])
                if (strncmp(str, type[i],3) == 0)
                        return(0);      /* found a match, file to be ignored */
        return(1);                      /* no match found */
}

FILE *
openfile(fn)
unsigned char	*fn;
{
	char	ans[80];
	unsigned char	*p, *q, s, *ext;
	int	procf;

	procf = TRUE;

	/* convert the VMS to UNIX and make the directory path */
	for (p = ufn, q = fn; *q != '\0' && *q != ']'; q++) {
	  if (*q == '[')
	    continue;
	  else if (*q == '.') {
	    *p = '\0';
	    if (procf && dflag) mkdir(ufn, 0777);
	    *p++ = '/';
	  } else
	    *p++ = isupper(*q) ? tolower(*q) : *q;
	}
	*p = '\0';
	if (*q == ']') q++;
	if (dflag) {
	  if (procf) {
	    mkdir(ufn, 0777);
	    *p++ = '/';
	  }
	} else p = ufn;

	for (; *q != '\0' && *q != ';'; q++) {
		*p++ = isupper(*q) ? tolower(*q) : *q;
		if ( *q == '.') ext = p;
	}
	if (cflag) {
		*p = ':';
		while (*q != '\0')
		  *p++ = *q++;
	}
	*p = '\0';

	if (!eflag && procf) procf = typecmp(ext);
	if (procf && wflag) {
		printf("extract %s [ny]", filename);
		fflush(stdout);
		gets(ans);
		if (*ans != 'y') procf = FALSE;
	}
	if (procf)
		/* open the file for writing */
		return(fopen(ufn, "w"));
	else
		return((FILE *)NULL);
}

/*
 *
 *  do swapping for BIGENDIAN architectures or copying for LITTLEENDIAN.
 *
 */
void
swap(from, to, nbytes)
     unsigned char *from, *to;
     int nbytes;
{
#ifdef	SWAP
  int i, j;
  unsigned char	temp;

  for (i = 0, j = nbytes-1; i <= j; i++, j--) {
    temp = from[i];
    to[i] = from[j];
    to[j] = temp;
  }
#else
  bcopy(from, to, nbytes);
#endif
}

/*
 *  Convert from VMS's 64 bit 100 nanosecond time to Unix's 32 bit 1 second
 *  time.
 */

time_t
ConvertVMSToUnixTime(VMSTime)
     unsigned char *VMSTime;
{
  MINT *mVMSTime, *mUnixEpochTime, *mChunkToSeconds, *mUnixTime;
  MINT *mTempMINT, *mTempMINT2;
  time_t unixTime;
  struct tm *t;
  unsigned char hex[17], *cp;

  /* Initialize the multiple precision integers */

  sprintf(hex, "%02x%02x%02x%02x%02x%02x%02x%02x",
	  VMSTime[7], VMSTime[6], VMSTime[5], VMSTime[4],
	  VMSTime[3], VMSTime[2], VMSTime[1], VMSTime[0]);
  mVMSTime = xtom(hex);
  mUnixEpochTime = xtom("007c95674beb4000"); /* 1 January 1970 00:00:00 */
  mChunkToSeconds = xtom("0000000000989680"); /* 10,000,000 */
  mUnixTime = itom(0);
  mTempMINT = itom(0);
  mTempMINT2 = itom(0);

  /* Do the arithmatic */

  if (mcmp(mVMSTime, mUnixEpochTime) >= 0)
    msub(mVMSTime, mUnixEpochTime, mTempMINT);
  mdiv(mTempMINT, mChunkToSeconds, mUnixTime, mTempMINT2);
  cp = (unsigned char *)mtox(mUnixTime);
  sscanf(cp, "%x", &unixTime);

  /* Free the multiple precision integers */

  mfree(mVMSTime);
  mfree(mUnixEpochTime);
  mfree(mUnixTime);
  mfree(mTempMINT);
  mfree(mTempMINT2);
  mfree(mChunkToSeconds);

  /* Convert from local time to GMT */

  t = gmtime(&unixTime);
#if defined(sun)
  unixTime = timelocal(t);
#elif defined(ultrix)
  unixTime = mktime(t);
#endif

  return (unixTime);
}

void
process_file(buffer, rsize)
     unsigned char *buffer;
     int rsize;
{
  int	i, n, procf, total_sizes;
  unsigned char	*p, *q;
  short	nblk, lnch, s;

  /* Close the old file and reset the counters. */

  CloseFile();

  swap(buffer, &s, sizeof(short));
  buffer += sizeof(short);
  rsize -= sizeof(short);

  /* check the header word */
  if (s != 257) {
    fprintf(stderr, "Snark: invalid data header\n");
    exit(1);
  }

  for (i = 0; i < FANO && rsize > sizeof(short) * 2 + 1; i++) {
    swap(buffer, &file_table[i].bsa_dol_w_size, sizeof(short));
    buffer += sizeof(short);
    rsize -= sizeof(short);

    swap(buffer, &file_table[i].bsa_dol_w_type, sizeof(short));
    buffer += sizeof(short);
    rsize -= sizeof(short);

    file_table[i].bsa_dol_t_text = buffer;
    buffer += file_table[i].bsa_dol_w_size;
    rsize -= file_table[i].bsa_dol_w_size;

#ifdef DEBUG
    if (debugFlag)
      printf("(%d) size %d, type %d rsize %d\n",
	     i, file_table[i].bsa_dol_w_size, file_table[i].bsa_dol_w_type,
	     rsize);
#endif /* DEBUG */

    switch (file_table[i].bsa_dol_w_type) {
    case 42:
      /* extract file name */
      p = file_table[i].bsa_dol_t_text;
      q = filename;
      for (n = 0; n < file_table[i].bsa_dol_w_size; n++)
	*q++ = *p++;
      *q = '\0';
      break;
    case 52:
      /* extract file's record attributes */
      p = file_table[i].bsa_dol_t_text;
      recfmt = p[0];
      recatt = p[1];
      swap(&p[2], &recsize, sizeof(short));
      vfcsize = p[15];
      if (vfcsize == 0)
	vfcsize = 2;
#ifdef DEBUG
      if (debugFlag) {
	printf("recfmt = %d\n", recfmt);
	printf("recatt = %d\n", recatt);
	printf("reclen = %d\n", recsize);
	printf("vfcsize = %d\n", vfcsize);
      }
#endif /* DEBUG */
      swap(&p[10], &nblk, sizeof(short));
      swap(&p[12], &lnch, sizeof(short));
      filesize = (nblk-1)*512 + lnch;
#ifdef DEBUG
      if (debugFlag) {
	printf("nbk = %d, lnch = %d\n", nblk, lnch);
	printf("filesize = 0x%x\n", filesize);
      }
#endif /* DEBUG */
      break;
    case 54:
    case 55:
    case 56:
    case 57:
      unixTimes[file_table[i].bsa_dol_w_type - 54] =
	ConvertVMSToUnixTime(file_table[i].bsa_dol_t_text);
#ifdef DEBUG
      if (debugFlag)
	printf("time %s",
	       ctime(&unixTimes[file_table[i].bsa_dol_w_type - 54]));
#endif /* DEBUG */
      break;
    }
  }

  /* open the file */
  procf = 0;
  if (goptind < gargc) 
    for (i = goptind; !procf && i < gargc; i++) {
      procf = match(filename,gargv[i]);
    }
  else
    procf = 1;
  if (tflag && procf) 
    printf( " %-35s %8d \n", filename, filesize);
  if (xflag && procf) {
    /* open file */
    f = openfile(filename);
    if (f != NULL && vflag)
      printf("extracting %s as %s\n", filename, ufn);
  }
}
/*
 *
 *  process a virtual block record (file record)
 *
 */
void
process_vbn(buffer, rsize)
char		*buffer;
unsigned int	rsize;
{
  int	c, i;

  if (f == NULL)
    return;

  i = 0;
  switch (recfmt) {
  case FAB_dol_C_FIX:
    while (file_count+i < filesize && i < rsize) {
      if (reclen == 0)
	reclen = recsize;

      fputc(buffer[i], f);
      i++;
      reclen--;
    }
    break;

  case FAB_dol_C_VAR:
  case FAB_dol_C_VFC:
    while (file_count+i < filesize && i < rsize) {
      if (reclen == 0) {
	swap(&buffer[i], &reclen, sizeof(short));
	i += 2;
	if (recfmt == FAB_dol_C_VFC) {
	  i += vfcsize;
	  reclen -= vfcsize;
	}
      } else {
	fputc(buffer[i], f);
	i++;
	reclen--;
      }
      if (reclen == 0) {
	if ((recatt & (FAB_dol_V_CR | FAB_dol_V_FTN | FAB_dol_V_PRN)) != 0)
	  fputc('\n', f);
	if (i & 1)
	  i++;
      }
    }
    break;

  case FAB_dol_C_STM:
  case FAB_dol_C_STMLF:
    while (file_count+i < filesize && i < rsize) {
      fputc(buffer[i], f);
      i++;
    }
    break;

  case FAB_dol_C_STMCR:
    while (file_count+i < filesize && i < rsize) {
      c = buffer[i++];
      if (c == '\r')
	fputc('\n', f);
      else
	fputc(c, f);
    }
    break;

  default:
    CloseFile();
    unlink(ufn);
    fprintf(stderr, "Invalid record format = %d\n", recfmt);
    return;
  }
  file_count += i;
}

/*
 *
 *  process a backup block
 *
 */
process_block(block, blocksize)
char	*block;
int	blocksize;
{

	unsigned short	bhsize, rsize, rtype;
	unsigned long	bsize, i;

	i = 0;

	/* read the backup block header */
	block_header = (struct bbh *) &block[i];
	i += sizeof(struct bbh);

	swap(&block_header->bbh_dol_w_size, &bhsize, sizeof(short));
	swap(&block_header->bbh_dol_l_blocksize, &bsize, sizeof(long));

	/* check the validity of the header block */
	if (bhsize != sizeof(struct bbh)) {
		fprintf(stderr, "Snark: Invalid header block size %d != %d\n",
			bhsize, sizeof(struct bbh));
		exit(1);
	}
	if (bsize != 0 && bsize != blocksize) {
		fprintf(stderr, "Snark: Invalid block size %d != %d\n",
			bsize, blocksize);
		exit(1);
	}
#ifdef DEBUG
	if (debugFlag)
	  printf("new block: i = %d, bsize = %d\n", i, bsize);
#endif /* DEBUG */

	/* read the records */
	while (i < bsize) {
		/* read the backup record header */
		record_header = (struct brh *) &block[i];
		i += sizeof(struct brh);

		swap(&record_header->brh_dol_w_rtype, &rtype, sizeof(short));
		swap(&record_header->brh_dol_w_rsize, &rsize, sizeof(short));
#ifdef DEBUG
		if (debugFlag) {
		  printf("rtype = %d\n", rtype);
		  printf("rsize = %d\n", rsize);
		  printf("flags = 0x%x\n", record_header->brh_dol_l_flags);
		  printf("addr = 0x%x\n", record_header->brh_dol_l_address);
		  printf("i = %d\n", i);
		}
#endif /* DEBUG */

		switch (rtype) {
		case brh_dol_k_null:
#ifdef DEBUG
		  if (debugFlag)
		    printf("rtype = null\n");
#endif /* DEBUG */
		  break;

		case brh_dol_k_summary:
#ifdef DEBUG
		  if (debugFlag)
		    printf("rtype = summary\n");
#endif /* DEBUG */
		  break;

		case brh_dol_k_volume:
#ifdef DEBUG
		  if (debugFlag)
		    printf("rtype = volume\n");
#endif /* DEBUG */
		  break;

		case brh_dol_k_file:
#ifdef DEBUG
		  if (debugFlag)
		    printf("rtype = file\n");
#endif /* DEBUG */
		  process_file(&block[i], (int)rsize);
		  break;

		case brh_dol_k_vbn:
#ifdef DEBUG
		  if (debugFlag)
		    printf("rtype = vbn\n");
#endif /* DEBUG */
		  process_vbn(&block[i], (int)rsize);
		  break;

		case brh_dol_k_physvol:
#ifdef DEBUG
		  if (debugFlag)
		    printf("rtype = physvol\n");
#endif /* DEBUG */
		  break;

		case brh_dol_k_lbn:
#ifdef DEBUG
		  if (debugFlag)
		    printf("rtype = lbn\n");
#endif /* DEBUG */
		  break;

		case brh_dol_k_fid:
#ifdef DEBUG
		  if (debugFlag)
		    printf("rtype = fid\n");
#endif /* DEBUG */
		  break;

		default:
		  fprintf(stderr, " Snark: invalid record type\n");
		  fprintf(stderr, " record type = %d\n", rtype);
		  exit(1);
		}
#ifdef pyr
		i = i + rsize;
#else
		i += rsize;
#endif
	}
}

int
rdhead()
{
  int i, nfound;
  char name[80];

  nfound = TRUE;
  /* read the tape label - 4 records of 80 bytes */
  if (istape) {
    while ((i = read(fd, label, LABEL_SIZE)) != 0) {
      if (i != LABEL_SIZE) {
	fprintf(stderr, "Snark: bad label record\n");
	exit(1);
      }
      if (strncmp(label, "VOL1",4) == 0) {
	sscanf(label+4, "%14s", name);
	if (vflag || tflag) printf("Volume: %s\n",name);
      }
      if (strncmp(label, "HDR1",4) == 0) {
	sscanf(label+4, "%14s", name);
	sscanf(label+31, "%4d", &setnr);
      }
      /* get the block size */
      if (strncmp(label, "HDR2", 4) == 0) {
	nfound = FALSE;
	sscanf(label+5, "%5d", &blocksize);
#ifdef DEBUG
	if (debugFlag)
	  printf("blocksize = %d\n", blocksize);
#endif /* DEBUG */
      }
    }
  } else {
    if (!bflag)
      blocksize = 32256;
    strncpy(name, tapefile, sizeof(name)-1);
    nfound = (setnr != 0);
    setnr = 1;
  }

  if ((vflag || tflag) && !nfound) 
    printf("Saveset name: %s   number: %d\n",name,setnr);
  /* get the block buffer */
  block = (char *) malloc(blocksize);
  if (block == (char *) 0) {
    fprintf(stderr, "memory allocation for block failed\n");
    exit(1);
  }
  return(nfound);
}

rdtail()
{
  int i;
  char name[80];

  if (istape) {
    /* read the tape label - 4 records of 80 bytes */
    while ((i = read(fd, label, LABEL_SIZE)) != 0) {
      if (i != LABEL_SIZE) {
	fprintf(stderr, "Snark: bad label record\n");
	exit(1);
      }
      if (strncmp(label, "EOF1",4) == 0) {
	sscanf(label+4, "%14s", name);
	if (vflag || tflag)
	  printf("End of saveset: %s\n\n\n",name);
      }
    }
  }
}

usage(progname)
char	*progname;
{
  fprintf(stderr,
	  "Usage:  %s -{tx}[Dcdevw][-b block size][-s setnumber][-f tapefile]\n",
	  progname);
}

main(argc, argv)
int	argc;
char	*argv[];
{
	
	char *progname;
	int	c, i, eoffl;
	int	selset;
	struct mtget mtstatus;
	extern int optind;
	extern char *optarg;

	progname = argv[0];
	if (argc < 2){
		usage(progname);
		exit(1);
	}
	gargv = argv;
	gargc = argc;
	tapefile = def_tapefile;
	istape = TRUE;
#ifdef DEBUG
	debugFlag = 0;
#endif /* DEBUG */
	bflag=cflag=dflag=eflag=sflag=tflag=vflag=wflag=xflag=0;
	while((c=getopt(argc,argv,"Db:cdef:s:tvwx")) != EOF)
		switch(c){
		case 'D':
#ifdef DEBUG
		  debugFlag++;
		  break;
#endif /* DEBUG */
		case 'b':
		        bflag++;
			sscanf(optarg, "%d", &blocksize);
			break;
		case 'c':
			cflag++;
			break;
		case 'd':
			dflag++;
			break;
		case 'e':
			eflag++;
			break;
		case 'f':
			tapefile = optarg;
			break;
		case 's':
			sflag++;
			sscanf(optarg,"%d",&selset);
			break;
		case 't':
			tflag++;
			break;
		case 'v':
			vflag++;
			break;
		case 'w':
			wflag++;
			break;
		case 'x':
			xflag++;
			break;
		case '?':
			usage(progname);
			exit(1);
			break;
		};
	if (!tflag && !xflag) {
		usage(progname);
		exit(1);
	}
	goptind = optind;

#ifdef	NEWD
	/* open debug file */
	lf = fopen("log", "w");
	if (lf == NULL) {
		perror("log");
		exit(1);
	}
#endif

	if (strcmp(tapefile, "-") == 0) {
	  fd = fileno(stdin);
	  istape = FALSE;
	} else {
	  /* open the tape file */
	  fd = open(tapefile, O_RDONLY);
	  if (fd < 0) {
	    perror(tapefile);
	    exit(1);
	  }
	}

	/* reading from a tape drive? */

	if (istape) {
	  i = ioctl(fd, MTIOCGET, &mtstatus);
	  if (i < 0) {
	    if (errno != ENOTTY) {
	      char errorString[256];

	      sprintf(errorString, "%s: MTIOCGET \"%s\"", progname, tapefile);
	      perror(errorString);
	      exit(1);
	    } else
	      istape = FALSE;
	  }
	}

	if (istape) {
	  op.mt_op = MTREW;	/* rewind tape */
	  op.mt_count = 1;
	  i = ioctl(fd, MTIOCTOP, &op);
	  if (i < 0) {
	    perror(tapefile);
	    exit(1);
	  }
	}

	eoffl = rdhead();
	/* read the backup tape blocks until end of tape */ 
	while (!eoffl) {
		if (sflag && setnr != selset) {
		  if (istape) {
		    op.mt_op = MTFSF;
		    op.mt_count = 1;
		    i = ioctl(fd, MTIOCTOP, &op);
		    if (i < 0) {
		      perror(tapefile);
		      exit(1);
		    }
		    i = 0;
		  }
		}
		else {
		  int bytesRead;

		  i = 0;
		  do {
		    bytesRead = read(fd, block + i, blocksize - i);
		    i += bytesRead;
		  } while (bytesRead > 0 && i < blocksize);
		}
		if (i == 0) {
		  rdtail();
		  eoffl=rdhead();
		}
		else if (i != blocksize) {
			fprintf(stderr, "bad block read i = %d\n", i);
			exit(1);
		}
		else {
			eoffl = 0;
			process_block(block, blocksize);
		}
	}
	if (vflag || tflag) {
	  if (istape)
	    printf("End of tape\n");
	  else
	    printf("End of saveset\n");
	}

	/* close the tape */
	close(fd);

	CloseFile();

#ifdef	NEWD
	/* close debug file */
	fclose(lf);
#endif	NEWD

	/* exit cleanly */
	exit(0);
}
