/*
 *
 *  Title:
 *	Backup
 *
 *  Decription:
 *	Program to read VMS backup tape
 *
 *  Author:
 *	John Douglas CAREY.
 *	Sven-Ove Westberg    (version 3.0)
 *	Pieter Bowman
 *
 *  Net-addess:
 *	john%monu1.oz@seismo.ARPA
 *	luthcad!sow@enea.UUCP
 *	bowman@math.utah.edu
 *
 *  History:
 *	Version 1.0 - September 1984
 *		Can only read variable length records
 *	Version 1.1
 *		Cleaned up the program from the original hack
 *		Can now read stream files
 *	Version 1.2
 *		Now convert filename from VMS to UNIX
 *			and creates sub-directories
 *	Version 1.3
 *		Works on the Pyramid if SWAP is defined
 *	Version 1.4
 *		Reads files spanning multiple tape blocks
 *	Version 1.5
 *		Always reset reclen = 0 on file open
 *		Now output fixed length records
 *
 *      Version 2.0 - July 1985
 *		VMS Version 4.0 causes a rethink !!
 *		Now use mtio operations instead of opening and closing file
 *		Blocksize now grabed from the label
 *
 *	Version 2.1 - September 1985
 *		Handle variable length records of zero length.
 *
 *	Version 2.2 - July 1986
 *		Handle FORTRAN records of zero length.
 *		Inserted exit(0) at end of program.
 *		Distributed program in aus.sources
 *
 *	Version 2.3 - August 1986
 *		Handle FORTRAN records with record length fields
 *		at the end of a block
 *		Put debug output to a file.
 *		Distributed program in net.sources
 *
 *	Version 3.0 - December 1986
 *		Handle multiple saveset
 *		Remote tape
 *		Interactive mode
 *		File name selection with meta-characters
 *		Convert ; to : in VMS filenames
 *		Flag for usage of VMS directory structure
 *		Flag for "useless" files  eg. *.exe
 *		Flag for use VMS version in file names
 *		Flag for verbose mode
 *		Flag to list the contents of the tape
 *		Distributed to mod.sources
 *
 *
 *		Changed code to allow reading from save set on disk.
 *		Added -b flag to handle changing block size.
 *
 *	Version 3.1 - August 1996
 *		Hacked a whole bunch since 1986
 *		Added pflag at the request of Bruce Lowery <blowery@utdallas.edu>
 *
 *  Installation:
 *
 *	Computer Centre
 *	Monash University
 *	Wellington Road
 *	Clayton
 *	Victoria	3168
 *	AUSTRALIA
 *
 */
#ifdef HAVE_CONFIG_H
#include	"config.h"
#endif /* HAVE_CONFIG_H */
#include	<stdio.h>
#include	<ctype.h>
#include	<errno.h>
#ifdef HAVE_SYS_IOCTL_H
#include	<sys/ioctl.h>
#endif /* HAVE_SYS_IOCTL_H */
#include	<sys/types.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif /* HAVE_STDLIB_H */

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif /* HAVE_MALLOC_H */

#ifdef HAS_TAPE
#ifdef REMOTE
#include	<local/rmt.h>
#endif /* REMOTE */
#endif /* HAS_TAPE */

#include	<sys/stat.h>
#include	<fcntl.h>

#ifdef HAS_TAPE
#ifdef HAVE_SYS_MTIO_H
#include	<sys/mtio.h>
#endif /*  HAVE_SYS_MTIO_H */
#endif /* HAS_TAPE */

#ifdef HAVE_FILE_H
#include	<sys/file.h>
#endif /*  HAVE_FILE_H */

#include	<time.h>

#define FALSE 0
#define TRUE 1

#ifndef UINT8
#define UINT8 u_char
#endif

#ifndef UINT16
#define UINT16 u_short
#endif

#ifndef UINT32
#define UINT32 u_long
#endif

struct bbh {
	UINT16	bbh_dol_w_size;
	UINT16	bbh_dol_w_opsys;
	UINT16	bbh_dol_w_subsys;
	UINT16	bbh_dol_w_applic;
	UINT32	bbh_dol_l_number;
	UINT8	bbh_dol_t_spare_1[20];
	UINT16	bbh_dol_w_struclev;
	UINT16	bbh_dol_w_volnum;
	UINT32	bbh_dol_l_crc;
	UINT32	bbh_dol_l_blocksize;
	UINT32	bbh_dol_l_flags;
	char	bbh_dol_t_ssname[32];
	UINT16	bbh_dol_w_fid[3];
	UINT16	bbh_dol_w_did[3];
	char	bbh_dol_t_filename[128];
	UINT8	bbh_dol_b_rtype;
	UINT8	bbh_dol_b_rattrib;
	UINT16	bbh_dol_w_rsize;
	UINT8	bbh_dol_b_bktsize;
	UINT8	bbh_dol_b_vfcsize;
	UINT16	bbh_dol_w_maxrec;
	UINT32	bbh_dol_l_filesize;
	UINT8	bbh_dol_t_spare_2[22];
	UINT16	bbh_dol_w_checksum;
};

struct brh {
	UINT16	brh_dol_w_rsize;
	UINT16	brh_dol_w_rtype;
	UINT32	brh_dol_l_flags;
	UINT32	brh_dol_l_address;
	UINT32	brh_dol_l_spare;
};

/* define record types */

#define	brh_dol_k_null	0
#define	brh_dol_k_summary	1
#define	brh_dol_k_volume	2
#define	brh_dol_k_file	3
#define	brh_dol_k_vbn	4
#define brh_dol_k_physvol	5
#define brh_dol_k_lbn	6
#define	brh_dol_k_fid	7

struct bsa {
	UINT16 bsa_dol_w_size;
	UINT16 bsa_dol_w_type;
	UINT8 *bsa_dol_t_text;
};

#ifdef HAS_TAPE
#ifdef	STREAM
char	*def_tapefile = "/dev/rts8";
#else /* STREAM */
char	*def_tapefile = "/dev/rmt8";
#endif /* STREAM */
#else /* HAS_TAPE */
char	*def_tapefile = NULL;
#endif /* HAS_TAPE */

char	*tapefile;

#ifdef HAS_TAPE
int istape;
#endif /* HAS_TAPE */

char	filename[128];	/* VMS file name */
char	ufn[256];	/* Unix file name */
UINT32	filesize;
time_t unixTimes[4];
UINT16 revisions;

UINT8	recfmt;			/* (bits 0-3) record format */
				/* (bits 4-7) file organization */

#define FAB_dol_C_UDF	0	/* undefined */
#define FAB_dol_C_FIX	1	/* fixed-length record */
#define FAB_dol_C_VAR	2	/* variable-length record */
#define FAB_dol_C_VFC	3	/* variable-length with fixed-length control record */
#define FAB_dol_C_STM	4	/* RMS stream record (valid only for sequential org) */
#define FAB_dol_C_STMLF	5	/* stream record delimited by LF (sequential org only) */
#define FAB_dol_C_STMCR	6	/* stream record delimited by CR (sequential org only) */
#define FAB_dol_C_MAXRFM	6	/* maximum rfm supported */

#define FAB_dol_C_SEQ 0		/* sequential */
#define FAB_dol_C_REL 16	/* relative */
#define FAB_dol_C_IDX 32	/* indexed */
#define FAB_dol_C_HSH 48	/* hashed */

UINT8	recatt;			/* record attributes */

#define FAB_dol_V_FTN	0x01	/* FORTRAN carriage control character */
#define FAB_dol_V_CR	0x02	/* line feed - record -carriage return */
#define FAB_dol_V_PRN	0x04	/* print-file carriage control */
#define FAB_dol_V_BLK	0x08	/* records don't cross block boundaries */

UINT16 recmax;
UINT16 extend;
UINT32 allocation;
UINT32 blocks;

FILE	*f	= NULL;
UINT32	file_count;
UINT16	reclen;
UINT16	recsize;
UINT32	vfcsize;

int	fd;		/* tape file descriptor */
#ifdef DEBUG
int 	debugFlag;
#endif /* DEBUG */
int 	bflag, cflag, dflag, eflag, pflag, sflag, tflag, vflag, wflag, xflag;
int	setnr = 0;		/* save set number */
char	**gargv;
int 	goptind, gargc;

#define	LABEL_SIZE	80
char	label[LABEL_SIZE];

UINT8	*block = (UINT8 *)0;
UINT32	blocksize;

#ifdef HAS_TAPE
struct	mtop	op;
#endif /* HAS_TAPE */

void
wordswap(from, to)
     UINT8 *from, *to;
{
  UINT8 c1, c2, c3, c4;

  c1 = *from++;
  c2 = *from++;
  c3 = *from++;
  c4 = *from++;
  *to++ = c3;
  *to++ = c4;
  *to++ = c1;
  *to++ = c2;
}

/*
 *
 *  do swapping for BIGENDIAN architectures or copying for LITTLEENDIAN.
 *
 */
void
swap(from, to, nbytes)
     UINT8 *from, *to;
     int nbytes;
{
#ifdef	SWAP
  int i, j;
  UINT8 temp;

  for (i = 0, j = nbytes-1; i <= j; i++, j--) {
    temp = from[i];
    to[i] = from[j];
    to[j] = temp;
  }
#else
  memcpy(to, from, nbytes);
#endif
}

void
dump_bbh(bbhp)
     struct bbh *bbhp;
{
  int i;
  UINT16	bbh_dol_w_size;
  short	bbh_dol_w_opsys;
  short	bbh_dol_w_subsys;
  short	bbh_dol_w_applic;
  UINT32	bbh_dol_l_number;
  short	bbh_dol_w_struclev;
  UINT16	bbh_dol_w_volnum;
  UINT32	bbh_dol_l_crc;
  UINT32	bbh_dol_l_blocksize;
  long	bbh_dol_l_flags;
  UINT16	bbh_dol_w_fid[3];
  short	bbh_dol_w_did[3];
  UINT16	bbh_dol_w_rsize;
  UINT8	bbh_dol_b_vfcsize;
  UINT16	bbh_dol_w_maxrec;
  UINT32	bbh_dol_l_filesize;
  UINT16	bbh_dol_w_checksum;

  swap((UINT8 *)&bbhp->bbh_dol_w_size,
       (UINT8 *)&bbh_dol_w_size,
       sizeof(bbhp->bbh_dol_w_size));

  swap((UINT8 *)&bbhp->bbh_dol_w_opsys,
       (UINT8 *)&bbh_dol_w_opsys,
       sizeof(bbhp->bbh_dol_w_opsys));

  swap((UINT8 *)&bbhp->bbh_dol_w_subsys,
       (UINT8 *)&bbh_dol_w_subsys,
       sizeof(bbhp->bbh_dol_w_subsys));

  swap((UINT8 *)&bbhp->bbh_dol_w_applic,
       (UINT8 *)&bbh_dol_w_applic,
       sizeof(bbhp->bbh_dol_w_applic));

  swap((UINT8 *)&bbhp->bbh_dol_l_number,
       (UINT8 *)&bbh_dol_l_number,
       sizeof(bbhp->bbh_dol_l_number));

  swap((UINT8 *)&bbhp->bbh_dol_w_struclev,
       (UINT8 *)&bbh_dol_w_struclev,
       sizeof(bbhp->bbh_dol_w_struclev));

  swap((UINT8 *)&bbhp->bbh_dol_w_volnum,
       (UINT8 *)&bbh_dol_w_volnum,
       sizeof(bbhp->bbh_dol_w_volnum));

  swap((UINT8 *)&bbhp->bbh_dol_l_crc,
       (UINT8 *)&bbh_dol_l_crc,
       sizeof(bbhp->bbh_dol_l_crc));

  swap((UINT8 *)&bbhp->bbh_dol_l_blocksize,
       (UINT8 *)&bbh_dol_l_blocksize,
       sizeof(bbhp->bbh_dol_l_blocksize));

  swap((UINT8 *)&bbhp->bbh_dol_l_flags,
       (UINT8 *)&bbh_dol_l_flags,
       sizeof(bbhp->bbh_dol_l_flags));

  swap((UINT8 *)&bbhp->bbh_dol_w_fid[0],
       (UINT8 *)&bbh_dol_w_fid[0],
       sizeof(bbhp->bbh_dol_w_fid[0]));

  swap((UINT8 *)&bbhp->bbh_dol_w_fid[1],
       (UINT8 *)&bbh_dol_w_fid[1],
       sizeof(bbhp->bbh_dol_w_fid[1]));

  swap((UINT8 *)&bbhp->bbh_dol_w_fid[2],
       (UINT8 *)&bbh_dol_w_fid[2],
       sizeof(bbhp->bbh_dol_w_fid[2]));

  swap((UINT8 *)&bbhp->bbh_dol_w_did[0],
       (UINT8 *)&bbh_dol_w_did[0],
       sizeof(bbhp->bbh_dol_w_did[0]));

  swap((UINT8 *)&bbhp->bbh_dol_w_did[1],
       (UINT8 *)&bbh_dol_w_did[1],
       sizeof(bbhp->bbh_dol_w_did[1]));

  swap((UINT8 *)&bbhp->bbh_dol_w_did[2],
       (UINT8 *)&bbh_dol_w_did[2],
       sizeof(bbhp->bbh_dol_w_did[2]));

  swap((UINT8 *)&bbhp->bbh_dol_w_rsize,
       (UINT8 *)&bbh_dol_w_rsize,
       sizeof(bbhp->bbh_dol_w_rsize));

  swap((UINT8 *)&bbhp->bbh_dol_w_maxrec,
       (UINT8 *)&bbh_dol_w_maxrec,
       sizeof(bbhp->bbh_dol_w_maxrec));

  swap((UINT8 *)&bbhp->bbh_dol_l_filesize,
       (UINT8 *)&bbh_dol_l_filesize,
       sizeof(bbhp->bbh_dol_l_filesize));

  swap((UINT8 *)&bbhp->bbh_dol_w_checksum,
       (UINT8 *)&bbh_dol_w_checksum,
       sizeof(bbhp->bbh_dol_w_checksum));

  printf("size = %d, opsys %d, subsys %d, applic %d, number %ld, spare_1",
	 bbh_dol_w_size,
	 bbh_dol_w_opsys,
	 bbh_dol_w_subsys,
	 bbh_dol_w_applic,
	 bbh_dol_l_number);
  for (i = 0; i < 20; i++)
    printf(" %u/0%o/0x%x",
	   bbhp->bbh_dol_t_spare_1[i],
	   bbhp->bbh_dol_t_spare_1[i],
	   bbhp->bbh_dol_t_spare_1[i]);
  printf(", struclev %d, volnum %d, crc 0x%lx, blocksize %d, flags 0x%x, ssname \"",
	 bbh_dol_w_struclev,
	 bbh_dol_w_volnum,
	 (unsigned long)bbh_dol_l_crc,
	 bbh_dol_l_blocksize,
	 bbh_dol_l_flags);
  for (i = 1; i <= bbhp->bbh_dol_t_ssname[0]; i++)
    printf("%c",
	   bbhp->bbh_dol_t_ssname[i]);
  printf("\", fid %d,%d,%d, did %d,%d,%d, filename \"",
	 bbh_dol_w_fid[0],
	 bbh_dol_w_fid[1],
	 bbh_dol_w_fid[2],
	 bbh_dol_w_did[0],
	 bbh_dol_w_did[1],
	 bbh_dol_w_did[2]);
  for (i = 1; i <= bbhp->bbh_dol_t_filename[0]; i++)
    printf("%c",
	   bbhp->bbh_dol_t_filename[i]);
  printf("\", rtype %d, rattrib %d, rsize %d, bktsize %d, vfcsize %d, maxrec %d, filesize %lu, spare_2",
	 bbhp->bbh_dol_b_rtype,
	 bbhp->bbh_dol_b_rattrib,
	 bbh_dol_w_rsize,
	 bbhp->bbh_dol_b_bktsize,
	 bbhp->bbh_dol_b_vfcsize,
	 bbh_dol_w_maxrec,
	 bbh_dol_l_filesize);
  for (i = 0; i < 22; i++)
    printf(" %u/0%o/0x%x",
	   bbhp->bbh_dol_t_spare_2[i],
	   bbhp->bbh_dol_t_spare_2[i],
	   bbhp->bbh_dol_t_spare_2[i]);
  printf(", checksum 0x%x\n",
	 bbh_dol_w_checksum);
}

char *
convert_time(t)
     time_t *t;
{
  static char ct_str[64];
  struct tm *tim;
  static char *month[] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
  };

  tim = localtime(t);
  sprintf(ct_str, "%2d-%s-%d %02d:%02d", tim->tm_mday, month[tim->tm_mon],
	  tim->tm_year, tim->tm_hour, tim->tm_min);
  return (ct_str);
}

void
CloseFile()
{
  time_t t2[2];

  if (f != NULL) {
    fclose(f);
    f = NULL;

    t2[0] = 0;			/* Never been accessed */
    t2[1] = unixTimes[1];	/* Use the revised date from VMS */
				/* for the Unix modify date. */
    utime(ufn, t2);
  }

  file_count = 0;
  reclen = 0;
}

int
typecmp(str)    /* Compare the filename type in str with our list
                   of file type to be ignored.  Return 0 if the
                   file is to be ignored, return 1 if the
                   file is not in our list and should not be ignored. */
register char   *str;
{
        static char *type[] = {
                "exe",          /* vms executable image */
                "lib",          /* vms object library */
                "obj",          /* rsx object file */
                "odl",          /* rsx overlay description file */
                "olb",          /* rsx object library */
                "pmd",          /* rsx post mortem dump */
                "stb",          /* rsx symbol table */
                "sys",          /* rsx bootable system image */
                "tsk",          /* rsx executable image */
		"dir",
		"upd",
		"tlo",
		"tlb",
                ""              /* null string terminates list */
        };
        register int    i;

        i = -1;
        while (*type[++i])
                if (strncmp(str, type[i],3) == 0)
                        return(0);      /* found a match, file to be ignored */
        return(1);                      /* no match found */
}

FILE *
openfile(fn)
char	*fn;
{
	char	ans[80];
	char	*p, *q, s, *ext;
	int	procf;

	procf = TRUE;

	/* convert the VMS to UNIX and make the directory path */
	for (p = ufn, q = fn; *q != '\0' && *q != ']'; q++) {
	  if (*q == '[')
	    continue;
	  else if (*q == '.') {
	    *p = '\0';
	    if (procf && dflag) mkdir(ufn, 0777);
	    *p++ = '/';
	  } else
	    *p++ = isupper(*q) ? tolower(*q) : *q;
	}
	*p = '\0';
	if (*q == ']') q++;
	if (dflag) {
	  if (procf) {
	    mkdir(ufn, 0777);
	    *p++ = '/';
	  }
	} else p = ufn;

	for (; *q != '\0' && *q != ';'; q++) {
		*p++ = isupper(*q) ? tolower(*q) : *q;
		if ( *q == '.') ext = p;
	}
	if (cflag && *q == ';') {
		q++;
		*p++ = ':';
		while (*q != '\0')
		  *p++ = *q++;
	}
	*p = '\0';

	if (!eflag && procf) procf = typecmp(ext);
	if (procf && wflag) {
		printf("extract %s [ny]", filename);
		fflush(stdout);
		gets(ans);
		if (*ans != 'y') procf = FALSE;
	}
	if (procf)
		/* open the file for writing */
		return(fopen(ufn, "w"));
	else
		return((FILE *)NULL);
}

/*
 *  Convert from VMS's 64 bit 100 nanosecond time to Unix's 32 bit 1 second
 *  time.
 */

time_t
ConvertVMSToUnixTime(VMSTime)
     UINT8 *VMSTime;
{
  time_t unixTime;
  struct tm *t;

#ifdef HAVE_LONGLONG
  long long llVMSTime;

  swap((UINT8 *)VMSTime, (UINT8 *)&llVMSTime, sizeof(llVMSTime));
  if (llVMSTime >= 35067168000000000) {
    unixTime = (time_t)((llVMSTime - 35067168000000000) / 10000000);
  } else {
    unixTime = 0;
  }
#else /* HAVE_LONGLONG */
#ifdef HAVE_MP

#include	<mp.h>

  MINT *mVMSTime, *mUnixEpochTime, *mChunkToSeconds, *mUnixTime;
  MINT *mTempMINT, *mTempMINT2;
  char hex[17], *cp;

  /* Initialize the multiple precision integers */

  sprintf(hex, "%02x%02x%02x%02x%02x%02x%02x%02x",
	  VMSTime[7], VMSTime[6], VMSTime[5], VMSTime[4],
	  VMSTime[3], VMSTime[2], VMSTime[1], VMSTime[0]);
  mVMSTime = xtom(hex);
  mUnixEpochTime = xtom("007c95674beb4000"); /* 1 January 1970 00:00:00 */
  mChunkToSeconds = xtom("0000000000989680"); /* 10,000,000 */
  mUnixTime = itom(0);
  mTempMINT = itom(0);
  mTempMINT2 = itom(0);

  /* Do the arithmatic */

  if (mcmp(mVMSTime, mUnixEpochTime) >= 0)
    msub(mVMSTime, mUnixEpochTime, mTempMINT);
  mdiv(mTempMINT, mChunkToSeconds, mUnixTime, mTempMINT2);
  cp = mtox(mUnixTime);
  sscanf(cp, "%x", &unixTime);
  fprintf(stderr, "hex %s, cp %s, unixTime %d\n", hex, cp, unixTime);
  free(cp);

  /* Free the multiple precision integers */

cp = mtox(mVMSTime) ; printf("  mVMSTime %s\n", cp); free(cp);
cp = mtox(mUnixEpochTime) ; printf("  mUnixEpochTime %s\n", cp); free(cp);
cp = mtox(mUnixTime) ; printf("  mUnixTime %s\n", cp); free(cp);
cp = mtox(mTempMINT) ; printf("  mTempMINT %s\n", cp); free(cp);
cp = mtox(mTempMINT2) ; printf("  mTempMINT2 %s\n", cp); free(cp);
cp = mtox(mChunkToSeconds) ; printf("  mChunkToSeconds %s\n", cp); free(cp);

  mfree(mVMSTime);
  mfree(mUnixEpochTime);
  mfree(mUnixTime);
  mfree(mTempMINT);
  mfree(mTempMINT2);
  mfree(mChunkToSeconds);

#else /* HAVE_MP */
  unixTime = time(NULL);
  return(unixTime);
#endif /* HAVE_MP */
#endif /* HAVE_LONGLONG */
  /* Convert from local time to GMT */

  t = gmtime(&unixTime);
  t->tm_isdst = -1;
#if defined(HAS_TIMELOCAL)
  unixTime = timelocal(t);
#elif defined(HAS_MKTIME)
  unixTime = mktime(t);
#endif

  return (unixTime);
}

void
process_file(buffer, rsize)
     UINT8 *buffer;
     int rsize;
{
  int i, procf, psize;
  UINT8 *p, *q;
  UINT16 eofbyte, gbc, dvl, s;
  UINT32 eofblock;
  UINT8 structure_level, structure_version, bucket_size;
  struct bsa bsad;

  /* Close the old file and reset the counters. */

  CloseFile();

  swap(buffer, (UINT8 *)&s, sizeof(s));
  buffer += sizeof(s);
  rsize -= sizeof(s);

  /* check the header word */
  if (s != 257) {
    fprintf(stderr, "Snark: invalid data header\n");
    exit(1);
  }

  while (rsize > sizeof(bsad.bsa_dol_w_size)+sizeof(bsad.bsa_dol_w_type)+1) {
    swap(buffer, (UINT8 *)&bsad.bsa_dol_w_size, sizeof(bsad.bsa_dol_w_size));
    buffer += sizeof(bsad.bsa_dol_w_size);
    rsize -= sizeof(bsad.bsa_dol_w_size);

    swap(buffer, (UINT8 *)&bsad.bsa_dol_w_type, sizeof(bsad.bsa_dol_w_type));
    buffer += sizeof(bsad.bsa_dol_w_type);
    rsize -= sizeof(bsad.bsa_dol_w_type);

    bsad.bsa_dol_t_text = buffer;
    buffer += bsad.bsa_dol_w_size;
    rsize -= bsad.bsa_dol_w_size;

#ifdef DEBUG
    if (debugFlag)
      printf("type %d size %d rsize %d",
	     bsad.bsa_dol_w_type,
	     bsad.bsa_dol_w_size,
	     rsize);
#endif /* DEBUG */

    p = bsad.bsa_dol_t_text;
    psize = (int)bsad.bsa_dol_w_size;

    switch (bsad.bsa_dol_w_type) {
    case 42:
      /* extract file name */
      for (q = (UINT8 *)filename; psize > 0;  psize--)
	*q++ = *p++;
      *q = '\0';
#ifdef DEBUG
      if (debugFlag) {
	printf(" Filename: %s", filename);
      }
#endif /* DEBUG */
      break;
    case 43:
      structure_version = *p++;
      psize--;
      structure_level = *p++;
      psize--;
#ifdef DEBUG
      if (debugFlag) {
	printf(" structure level and version: %d, %d",
	       structure_level, structure_version);
      }
#endif /* DEBUG */
      break;
    case 44:
#ifdef DEBUG
      if (debugFlag) {
	UINT16 file_id[3];

	swap(p, (UINT8 *)&file_id[0], sizeof(UINT16));
	p += sizeof(UINT16);
	swap(p, (UINT8 *)&file_id[1], sizeof(UINT16));
	p += sizeof(UINT16);
	swap(p, (UINT8 *)&file_id[2], sizeof(UINT16));
	p += sizeof(UINT16);
	printf(" File ID: (%u,%u,%u)",
	       file_id[0],
	       file_id[1],
	       file_id[2]);
      }
#endif /* DEBUG */
      break;
    case 45:
#ifdef DEBUG
      if (debugFlag) {
	UINT16 file_id[3];

	swap(p, (UINT8 *)&file_id[0], sizeof(UINT16));
	p += sizeof(UINT16);
	swap(p, (UINT8 *)&file_id[1], sizeof(UINT16));
	p += sizeof(UINT16);
	swap(p, (UINT8 *)&file_id[2], sizeof(UINT16));
	p += sizeof(UINT16);
	printf(" Back link file identification: (%d,%d,%d)",
	       file_id[0],
	       file_id[1],
	       file_id[2]);
      }
#endif /* DEBUG */
      break;
    case 46:
      swap(p, (UINT8 *)&blocks, sizeof(blocks));
      p += sizeof(blocks);
      psize -= sizeof(blocks);
#ifdef DEBUG
      if (debugFlag) {
	printf(" Blocks: %u",
	       blocks);
      }
#endif /* DEBUG */
      break;
    case 47:
#ifdef DEBUG
      if (debugFlag) {
	UINT16 uic[2];

	swap(p, (UINT8 *)&uic[0], sizeof(UINT16));
	p += sizeof(UINT16);
	swap(p, (UINT8 *)&uic[1], sizeof(UINT16));
	p += sizeof(UINT16);
	printf(" Owner: [%06o,%06o]",
	       uic[1],
	       uic[0]);
      }
#endif /* DEBUG */
      break;
    case 48:
#ifdef DEBUG
      if (debugFlag) {
	printf(" (u_char)");
	for (; psize > 0; psize--, p++)
	  printf(" %lu/0%lo/0x%lx",
		 (UINT32)*p, (UINT32)*p, (UINT32)*p);
      }
#endif /* DEBUG */
      break;
    case 49:
#ifdef DEBUG
      if (debugFlag) {
	printf(" (u_char)");
	for (; psize > 0; psize--, p++)
	  printf(" %lu/0%lo/0x%lx",
		 (UINT32)*p, (UINT32)*p, (UINT32)*p);
      }
#endif /* DEBUG */
      break;
    case 50:
#ifdef DEBUG
      if (debugFlag) {
	printf(" (u_char)");
	for (; psize > 0; psize--, p++)
	  printf(" %lu/0%lo/0x%lx",
		 (UINT32)*p, (UINT32)*p, (UINT32)*p);
      }
#endif /* DEBUG */
      break;
    case 51:
#ifdef DEBUG
      if (debugFlag) {
	UINT32 c;

	swap(p, (UINT8 *)&c, sizeof(c));
	printf(" file characteristics:");
	if (c == 0) {
	  printf(" none");
	}
	if ((c & 0x80) != 0) {
	  printf(" contiguous");
	}
	if ((c & 0x20) != 0) {
	  printf(" contiguous best try");
	}
	if ((c & 0x2000) != 0) {
	  printf(" directory");
	}
	printf(" (u_long) %lu/0%lo/0x%lx",
	       c, c, c);
	p += sizeof(UINT32);
      }
#endif /* DEBUG */
      break;
    case 52:
      /* extract file's record attributes */
      recfmt = *p++;
      psize--;
      recatt = *p++;
      psize--;
      swap(p, (UINT8 *)&recsize, sizeof(recsize)); /* Record size */
      p += sizeof(recsize);
      psize -= sizeof(recsize);
      wordswap(p, (UINT8 *)&allocation);
      swap((UINT8 *)&allocation, (UINT8 *)&allocation, sizeof(allocation)); /* Highest block */
      p += sizeof(allocation);
      psize -= sizeof(allocation);
      wordswap(p, (UINT8 *)&eofblock);
      swap((UINT8 *)&eofblock, (UINT8 *)&eofblock, sizeof(eofblock)); /* End of file block */
      p += sizeof(eofblock);
      psize -= sizeof(eofblock);
      swap(p, (UINT8 *)&eofbyte, sizeof(eofbyte)); /* End of file byte */
      p += sizeof(eofbyte);
      psize -= sizeof(eofbyte);
      filesize = (allocation - (UINT32)1)*(UINT32)512 + (UINT32)eofbyte;
      bucket_size = *p++;	/* Bucket size */
      psize--;
      vfcsize = *p++;		/* Fixed control area size */
      psize--;
      if (vfcsize == 0)
	vfcsize = 2;
      swap(p, (UINT8 *)&recmax, sizeof(recmax)); /* Maximum record size */
      p += sizeof(recmax);
      psize -= sizeof(recmax);
      swap(p, (UINT8 *)&extend, sizeof(extend)); /* Default extension size */
      p += sizeof(extend);
      psize -= sizeof(extend);
      swap(p, (UINT8 *)&gbc, sizeof(gbc)); /* Global buffer count */
      p += sizeof(gbc);
      psize -= sizeof(gbc);
      swap(p, (UINT8 *)&dvl, sizeof(dvl)); /* Directory version limit */
      p += sizeof(dvl);
      psize -= sizeof(dvl);
#ifdef DEBUG
      if (debugFlag) {
	printf(" (u_char)");
	for (; psize > 0; psize--, p++)
	  printf(" %lu/0%lo/0x%lx",
		 (UINT32)*p, (UINT32)*p, (UINT32)*p);
	printf(" recfmt = %d, recatt = %d, recsize = %ld, allocation = %ld, eofblock = %lu, eofbyte = %lu, bucket_size = %d, vfcsize = %ld, recmax = %d, extend = %d, gbc = %d, dvl = %d, filesize = %lu",
	       recfmt,
	       recatt,
	       (UINT32)recsize,
	       allocation,
	       (UINT32)eofblock,
	       (UINT32)eofbyte,
	       bucket_size,
	       vfcsize,
	       recmax,
	       extend,
	       gbc,
	       dvl,
	       filesize);
      } else {
	p += psize;
	psize -= psize;
      }
#else
      p += psize;
      psize -= psize;
#endif /* DEBUG */
      break;
    case 53:
      swap(p, (UINT8 *)&revisions, sizeof(revisions));
      p += sizeof(revisions);
      psize -= sizeof(revisions);

#ifdef DEBUG
      if (debugFlag) {
	printf(" Revisions: %lu",
	       (UINT32)revisions);
      }
#endif /* DEBUG */
      break;
    case 54:
    case 55:
    case 56:
    case 57:
      unixTimes[bsad.bsa_dol_w_type - 54] =
	ConvertVMSToUnixTime(p);
#ifdef DEBUG
      if (debugFlag)
	printf(" time %s",
	       convert_time(&unixTimes[bsad.bsa_dol_w_type - 54]));
#endif /* DEBUG */
      break;
    case 71:
#ifdef DEBUG
      if (debugFlag) {
	printf(" (u_char)");
	for (; psize > 0; psize--, p++)
	  printf(" %lu/0%lo/0x%lx",
		 (UINT32)*p, (UINT32)*p, (UINT32)*p);
      }
#endif /* DEBUG */
      break;
    case 72:
#ifdef DEBUG
      if (debugFlag) {
	printf(" (u_char)");
	for (; psize > 0; psize--, p++)
	  printf(" %lu/0%lo/0x%lx",
		 (UINT32)*p, (UINT32)*p, (UINT32)*p);
      }
#endif /* DEBUG */
      break;
    case 73:
#ifdef DEBUG
      if (debugFlag) {
	printf(" (u_char)");
	for (; psize > 0; psize--, p++)
	  printf(" %lu/0%lo/0x%lx",
		 (UINT32)*p, (UINT32)*p, (UINT32)*p);
      }
#endif /* DEBUG */
      break;
    case 74:
#ifdef DEBUG
      if (debugFlag) {
	printf(" (u_char)");
	for (; psize > 0; psize--, p++)
	  printf(" %lu/0%lo/0x%lx",
		 (UINT32)*p, (UINT32)*p, (UINT32)*p);
      }
#endif /* DEBUG */
      break;
    case 75:
#ifdef DEBUG
      if (debugFlag) {
	printf(" (u_char)");
	for (; psize > 0; psize--, p++)
	  printf(" %lu/0%lo/0x%lx",
		 (UINT32)*p, (UINT32)*p, (UINT32)*p);
      }
#endif /* DEBUG */
      break;
    case 79:
#ifdef DEBUG
      if (debugFlag) {
	printf(" (u_char)");
	for (; psize > 0; psize--, p++)
	  printf(" %lu/0%lo/0x%lx",
		 (UINT32)*p, (UINT32)*p, (UINT32)*p);
      }
#endif /* DEBUG */
      break;
    case 80:
#ifdef DEBUG
      if (debugFlag) {
	printf(" (u_char)");
	for (; psize > 0; psize--, p++)
	  printf(" %lu/0%lo/0x%lx",
		 (UINT32)*p, (UINT32)*p, (UINT32)*p);
      }
#endif /* DEBUG */
      break;
    case 87:
#ifdef DEBUG
      if (debugFlag) {
	printf(" (u_char)");
	for (; psize > 0; psize--, p++)
	  printf(" %lu/0%lo/0x%lx",
		 (UINT32)*p, (UINT32)*p, (UINT32)*p);
      }
#endif /* DEBUG */
      break;
    }
#ifdef DEBUG
    if (debugFlag) {
      printf("\n");
    }
#endif /* DEBUG */
  }

  /* open the file */
  procf = 0;
  if (goptind < gargc)
    for (i = goptind; !procf && i < gargc; i++) {
      procf = match(filename,gargv[i]);
    }
  else
    procf = 1;
  if (tflag && procf)
    printf( " %-35s %8lu \n", filename, filesize);
  if (xflag && procf) {
    /* open file */
    f = openfile(filename);
    if (f != NULL && vflag)
      printf("extracting %s as %s\n", filename, ufn);
  }
}
/*
 *
 *  process a virtual block record (file record)
 *
 */
void
process_vbn(buffer, rsize)
UINT8		*buffer;
int	rsize;
{

  if (f == NULL)
    return;

  switch (recfmt & 0xf) {
  case FAB_dol_C_FIX:
    while (file_count < filesize && rsize > 0) {
      if (reclen == 0)
	reclen = recsize;

      fputc(*buffer, f);
      buffer++;
      file_count++;
      rsize--;
      reclen--;
    }
    break;

  case FAB_dol_C_VAR:
  case FAB_dol_C_VFC:
    {
      int i;

      i = 0;
      while (file_count < filesize && rsize > 0) {
	if (reclen == 0) {
	  swap(buffer, (UINT8 *)&reclen, sizeof(reclen));
	  if (reclen == 0xffff)
	    break;
	  if (pflag) {
	    fputc(*buffer, f);
	    fputc(*(buffer+1), f);
	  }
	  buffer += sizeof(reclen);
	  i += sizeof(reclen);
	  file_count += sizeof(reclen);
	  rsize -= sizeof(reclen);
	  if ((recfmt & 0xf) == FAB_dol_C_VFC) {
	    buffer += vfcsize;
	    i += vfcsize;
	    file_count += vfcsize;
	    rsize -= vfcsize;
	    reclen -= vfcsize;
	  }
	} else {
	  fputc(*buffer, f);
	  buffer++;
	  i++;
	  file_count++;
	  rsize--;
	  reclen--;
	}
	if (reclen == 0) {
	  if ((recatt & (FAB_dol_V_CR | FAB_dol_V_FTN | FAB_dol_V_PRN)) != 0)
	    fputc('\n', f);
	  if (i & 1) {
	    buffer++;
	    i++;
	    file_count++;
	    rsize--;
	  }
	}
      }
    }
    break;

  case FAB_dol_C_STM:
  case FAB_dol_C_STMLF:
    while (file_count < filesize && rsize > 0) {
      fputc(*buffer, f);
      buffer++;
      file_count++;
      rsize--;
    }
    break;

  case FAB_dol_C_STMCR:
    while (file_count < filesize && rsize > 0) {
      if (*buffer == '\r')
	fputc('\n', f);
      else
	fputc(*buffer, f);
      buffer++;
      file_count++;
      rsize--;
    }
    break;

  default:
    CloseFile();
    unlink(ufn);
    fprintf(stderr, "Invalid record format = %d\n", recfmt);
  }
}

/*
 *
 *  process a backup block
 *
 */
void
process_block(blockp, blocklength)
UINT8	*blockp;
int	blocklength;
{
  struct bbh *bbhp;
  struct brh *brhp;
  UINT16 bbh_size;
  UINT32 bbh_blocksize;
  UINT16 brh_rsize;
  UINT16 brh_rtype;


  /* read the backup block header */

  bbhp = (struct bbh *)blockp;
  blockp += sizeof(struct bbh);
  blocklength -= sizeof(struct bbh);

  swap((UINT8 *)&bbhp->bbh_dol_w_size,
       (UINT8 *)&bbh_size,
       sizeof(bbhp->bbh_dol_w_size));

  swap((UINT8 *)&bbhp->bbh_dol_l_blocksize,
       (UINT8 *)&bbh_blocksize,
       sizeof(bbhp->bbh_dol_l_blocksize));


  /* check the validity of the header block */

  if (bbh_size != sizeof(struct bbh)) {
    fprintf(stderr,
	    "Snark: Invalid header block size %d != %d\n",
	    bbh_size,
	    sizeof(struct bbh));
    exit(1);
  }
  if (bbh_blocksize != 0 && bbh_blocksize != blocklength + sizeof(struct bbh)) {
    fprintf(stderr,
	    "Snark: Invalid block size %d != %d\n",
	    bbh_blocksize,
	    blocklength + sizeof(struct bbh));
    exit(1);
  }
#ifdef DEBUG
  if (debugFlag)
    dump_bbh(bbhp);
#endif				/* DEBUG */

  if (bbh_blocksize == 0)
    return;

  /* read the records */

  while (blocklength > sizeof(struct brh)) {
    /* read the backup record header */

    brhp = (struct brh *)blockp;
    blockp += sizeof(struct brh);
    blocklength -= sizeof(struct brh);

    swap((UINT8 *)&brhp->brh_dol_w_rtype,
	 (UINT8 *)&brh_rtype,
	 sizeof(brhp->brh_dol_w_rtype));

    swap((UINT8 *)&brhp->brh_dol_w_rsize,
	 (UINT8 *)&brh_rsize,
	 sizeof(brhp->brh_dol_w_rsize));

#ifdef DEBUG
    if (debugFlag) {
      UINT32 brh_flags;
      UINT32 brh_address;
      UINT32 brh_spare;

      swap((UINT8 *)&brhp->brh_dol_l_flags,
	   (UINT8 *)&brh_flags,
	   sizeof(brhp->brh_dol_l_flags));

      swap((UINT8 *)&brhp->brh_dol_l_address,
	   (UINT8 *)&brh_address,
	   sizeof(brhp->brh_dol_l_address));

      swap((UINT8 *)&brhp->brh_dol_l_spare,
	   (UINT8 *)&brh_spare,
	   sizeof(brhp->brh_dol_l_spare));

      printf("blocklength = %d, rtype = %d, rsize = %d, flags = 0x%lx, addr = 0x%lx, spare = 0x%lx\n",
	     blocklength,
	     brh_rtype,
	     brh_rsize,
	     brh_flags,
	     brh_address,
	     brh_spare);
    }
#endif				/* DEBUG */

    switch (brh_rtype) {
    case brh_dol_k_null:
#ifdef DEBUG
      if (debugFlag)
	printf("rtype = null\n");
#endif				/* DEBUG */
      break;

    case brh_dol_k_summary:
#ifdef DEBUG
      if (debugFlag)
	printf("rtype = summary\n");
#endif				/* DEBUG */
      break;

    case brh_dol_k_volume:
#ifdef DEBUG
      if (debugFlag)
	printf("rtype = volume\n");
#endif				/* DEBUG */
      break;

    case brh_dol_k_file:
#ifdef DEBUG
      if (debugFlag)
	printf("rtype = file\n");
#endif				/* DEBUG */
      process_file(blockp, (int)brh_rsize);
      break;

    case brh_dol_k_vbn:
#ifdef DEBUG
      if (debugFlag)
	printf("rtype = vbn\n");
#endif				/* DEBUG */
      process_vbn(blockp, (int)brh_rsize);
      break;

    case brh_dol_k_physvol:
#ifdef DEBUG
      if (debugFlag)
	printf("rtype = physvol\n");
#endif				/* DEBUG */
      break;

    case brh_dol_k_lbn:
#ifdef DEBUG
      if (debugFlag)
	printf("rtype = lbn\n");
#endif				/* DEBUG */
      break;

    case brh_dol_k_fid:
#ifdef DEBUG
      if (debugFlag)
	printf("rtype = fid\n");
#endif				/* DEBUG */
      break;

    default:
      fprintf(stderr,
	      "Snark: invalid record type = %d\n",
	      brh_rtype);
      /*      exit(1); */
    }
    blockp += brh_rsize;
    blocklength -= brh_rsize;
  }
}

int
rdhead()
{
  int i, nfound;
  char name[80];

  nfound = TRUE;
  /* read the tape label - 4 records of 80 bytes */
#ifdef HAS_TAPE
  if (istape) {
    while ((i = read(fd, label, LABEL_SIZE)) != 0) {
      if (i != LABEL_SIZE) {
	fprintf(stderr, "Snark: bad label record\n");
	exit(1);
      }
      if (strncmp(label, "VOL1",4) == 0) {
	sscanf(label+4, "%14s", name);
	if (vflag || tflag) printf("Volume: %s\n",name);
      }
      if (strncmp(label, "HDR1",4) == 0) {
	sscanf(label+4, "%14s", name);
	sscanf(label+31, "%4d", &setnr);
      }
      /* get the block size */
      if (strncmp(label, "HDR2", 4) == 0) {
	nfound = FALSE;
	sscanf(label+5, "%5d", &blocksize);
#ifdef DEBUG
	if (debugFlag)
	  printf("blocksize = %d\n", blocksize);
#endif /* DEBUG */
      }
    }
  } else
#endif /* HAS_TAPE */
  {
    if (!bflag)
      blocksize = 32256;
    strncpy(name, tapefile, sizeof(name)-1);
    nfound = (setnr != 0);
    setnr = 1;
  }

  if ((vflag || tflag) && !nfound)
    printf("Saveset name: %s   number: %d\n",name,setnr);
  /* get the block buffer */
  if (block == (UINT8 *)0) {
    block = (UINT8 *) malloc(blocksize);
    if (block == (UINT8 *)0) {
      fprintf(stderr, "memory allocation for block failed\n");
      exit(1);
    }
  }
  return(nfound);
}

void
rdtail()
{
#ifdef HAS_TAPE
  int i;
  char name[80];

  if (istape) {
    /* read the tape label - 4 records of 80 bytes */
    while ((i = read(fd, label, LABEL_SIZE)) != 0) {
      if (i != LABEL_SIZE) {
	fprintf(stderr, "Snark: bad label record\n");
	exit(1);
      }
      if (strncmp(label, "EOF1",4) == 0) {
	sscanf(label+4, "%14s", name);
	if (vflag || tflag)
	  printf("End of saveset: %s\n\n\n",name);
      }
    }
  }
#endif /* HAS_TAPE */
}

void
usage(progname)
char	*progname;
{
#ifdef DEBUG
  fprintf(stderr,
	  "Usage:  %s -{tx}[Dcdepvw][b block size][s setnumber][f tapefile] [ name ... ]\n",
	  progname);
#else /* DEBUG */
  fprintf(stderr,
	  "Usage:  %s -{tx}[cdepvw][b block size][s setnumber][f tapefile] [ name ... ]\n",
	  progname);
#endif /* DEBUG */
  fprintf(stderr, "\t-t -- table of contents\n");
  fprintf(stderr, "\t-x -- extract files\n");
  fprintf(stderr, "\n");
#ifdef DEBUG
  fprintf(stderr, "\t-D -- debugging\n");
#endif /* DEBUG */
  fprintf(stderr, "\t-c -- use complete filenames, including the version number\n");
  fprintf(stderr, "\t-d -- use the directory structure from VMS\n");
  fprintf(stderr, "\t-e -- process all filename extensions\n");
  fprintf(stderr, "\t-p -- preserve 2-byte record headers in VAR and VFC files\n");
  fprintf(stderr, "\t-v -- verbose output\n");
  fprintf(stderr, "\t-w -- wait for user confirmation\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "\t-b -- use a different block size than the default\n");
  fprintf(stderr, "\t-s -- process only the given saveset number\n");
  fprintf(stderr, "\t-f -- tapefile (default %s) ('-' is standard input)\n", def_tapefile);
}

int
main(argc, argv)
int	argc;
char	*argv[];
{

	char *progname;
	int	c, i, eoffl;
	int	selset;
#ifdef HAS_TAPE
	struct mtget mtstatus;
#endif /* HAS_TAPE */
	extern int optind;
	extern char *optarg;

	progname = argv[0];
	if (argc < 2){
		usage(progname);
		exit(1);
	}
	gargv = argv;
	gargc = argc;
	tapefile = def_tapefile;
#ifdef HAS_TAPE
	istape = TRUE;
#endif /* HAS_TAPE */
#ifdef DEBUG
	debugFlag = 0;
#endif /* DEBUG */
	bflag=cflag=dflag=eflag=pflag=sflag=tflag=vflag=wflag=xflag=0;
	while((c=getopt(argc,argv,"Db:cdef:ps:tvwx")) != EOF)
		switch(c){
		case 'D':
#ifdef DEBUG
		  debugFlag++;
		  break;
#endif /* DEBUG */
		case 'b':
		        bflag++;
			sscanf(optarg, "%d", &blocksize);
			break;
		case 'c':
			cflag++;
			break;
		case 'd':
			dflag++;
			break;
		case 'e':
			eflag++;
			break;
		case 'f':
			tapefile = optarg;
			break;
		case 'p':
			pflag++;
			break;
		case 's':
			sflag++;
			sscanf(optarg,"%d",&selset);
			break;
		case 't':
			tflag++;
			break;
		case 'v':
			vflag++;
			break;
		case 'w':
			wflag++;
			break;
		case 'x':
			xflag++;
			break;
		case '?':
			usage(progname);
			exit(1);
			break;
		};
	if (!tflag && !xflag) {
		usage(progname);
		exit(1);
	}
	goptind = optind;

	if (strcmp(tapefile, "-") == 0) {
	  fd = fileno(stdin);
#ifdef HAS_TAPE
	  istape = FALSE;
#endif /* HAS_TAPE */
	} else {
	  /* open the tape file */
#ifdef O_BINARY
	  fd = open(tapefile, O_RDONLY|O_BINARY);
#else /* O_BINARY */
	  fd = open(tapefile, O_RDONLY);
#endif /* O_BINARY */
	  if (fd < 0) {
	    perror(tapefile);
	    exit(1);
	  }
	}

	/* reading from a tape drive? */

#ifdef HAS_TAPE
	if (istape) {
	  i = ioctl(fd, MTIOCGET, &mtstatus);
	  if (i < 0) {
	    if (errno != ENOTTY && errno != EOPNOTSUPP) {
	      char errorString[256];

	      sprintf(errorString, "%s: MTIOCGET \"%s\"", progname, tapefile);
	      perror(errorString);
	      exit(1);
	    } else
	      istape = FALSE;
	  }
	}

	if (istape) {
	  op.mt_op = MTREW;	/* rewind tape */
	  op.mt_count = 1;
	  i = ioctl(fd, MTIOCTOP, &op);
	  if (i < 0) {
	    perror(tapefile);
	    exit(1);
	  }
	}
#endif /*  HAS_TAPE */

	eoffl = rdhead();
	/* read the backup tape blocks until end of tape */
	while (!eoffl) {
		if (sflag && setnr != selset) {
#ifdef HAS_TAPE
		  if (istape) {
		    op.mt_op = MTFSF;
		    op.mt_count = 1;
		    i = ioctl(fd, MTIOCTOP, &op);
		    if (i < 0) {
		      perror(tapefile);
		      exit(1);
		    }
		    i = 0;
		  }
#endif /* HAS_TAPE */
		} else {
		  int bytesRead;

		  i = 0;
		  do {
		    bytesRead = read(fd, block + i, blocksize - i);
		    i += bytesRead;
		  } while (bytesRead > 0 && i < blocksize);
		}
		if (i == 0) {
		  rdtail();
		  eoffl=rdhead();
		}
		else if (i != blocksize) {
			fprintf(stderr, "bad block read, read %d, instead of %d\n", i, blocksize);
			exit(1);
		}
		else {
			eoffl = 0;
			process_block(block, blocksize);
		}
	}
	if (vflag || tflag) {
#ifdef HAS_TAPE
	  if (istape)
	    printf("End of tape\n");
	  else
#endif /* HAS_TAPE */
	    printf("End of saveset\n");
	}

	/* close the tape */
	close(fd);

	CloseFile();

	/* exit cleanly */
	exit(0);
	return(0);
}
